"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getInstructionInputDefaultFragment = void 0;
const nodes_1 = require("../../../nodes");
const shared_1 = require("../../../shared");
const visitors_1 = require("../../../visitors");
const asyncHelpers_1 = require("../asyncHelpers");
const common_1 = require("./common");
function getInstructionInputDefaultFragment(scope) {
    var _a, _b, _c, _d;
    const { input, optionalAccountStrategy, asyncResolvers, useAsync, nameApi, typeManifestVisitor, } = scope;
    if (!input.defaultValue) {
        return (0, common_1.fragment)('');
    }
    if (!useAsync && (0, asyncHelpers_1.isAsyncDefaultValue)(input.defaultValue, asyncResolvers)) {
        return (0, common_1.fragment)('');
    }
    const { defaultValue } = input;
    const defaultFragment = (renderedValue, isWritable) => {
        const inputName = (0, shared_1.camelCase)(input.name);
        if (input.kind === 'instructionAccountNode' &&
            (0, nodes_1.isNode)(defaultValue, 'resolverValueNode')) {
            return (0, common_1.fragment)(`accounts.${inputName} = { ...accounts.${inputName}, ...${renderedValue} };`);
        }
        if (input.kind === 'instructionAccountNode' && isWritable === undefined) {
            return (0, common_1.fragment)(`accounts.${inputName}.value = ${renderedValue};`);
        }
        if (input.kind === 'instructionAccountNode') {
            return (0, common_1.fragment)(`accounts.${inputName}.value = ${renderedValue};\n` +
                `accounts.${inputName}.isWritable = ${isWritable ? 'true' : 'false'}`);
        }
        return (0, common_1.fragment)(`args.${inputName} = ${renderedValue};`);
    };
    switch (defaultValue.kind) {
        case 'accountValueNode':
            const name = (0, shared_1.camelCase)(defaultValue.name);
            if (input.kind === 'instructionAccountNode' &&
                input.resolvedIsSigner &&
                !input.isSigner) {
                return defaultFragment(`expectTransactionSigner(accounts.${name}.value).address`).addImports('shared', 'expectTransactionSigner');
            }
            if (input.kind === 'instructionAccountNode') {
                return defaultFragment(`expectSome(accounts.${name}.value)`).addImports('shared', 'expectSome');
            }
            return defaultFragment(`expectAddress(accounts.${name}.value)`).addImports('shared', 'expectAddress');
        case 'pdaValueNode':
            const pdaFunction = nameApi.pdaFindFunction(defaultValue.pda.name);
            const pdaImportFrom = (_a = defaultValue.pda.importFrom) !== null && _a !== void 0 ? _a : 'generatedPdas';
            const pdaArgs = [];
            const pdaSeeds = defaultValue.seeds.map((seed) => {
                if ((0, nodes_1.isNode)(seed.value, 'accountValueNode')) {
                    return (0, common_1.fragment)(`${seed.name}: expectAddress(accounts.${(0, shared_1.camelCase)(seed.value.name)}.value)`).addImports('shared', 'expectAddress');
                }
                if ((0, nodes_1.isNode)(seed.value, 'argumentValueNode')) {
                    return (0, common_1.fragment)(`${seed.name}: expectSome(args.${(0, shared_1.camelCase)(seed.value.name)})`).addImports('shared', 'expectSome');
                }
                return (0, visitors_1.visit)(seed.value, typeManifestVisitor).value.mapRender((r) => `${seed.name}: ${r}`);
            });
            const pdaSeedsFragment = (0, common_1.mergeFragments)(pdaSeeds, (renders) => renders.join(', ')).mapRender((r) => `{ ${r} }`);
            if (pdaSeeds.length > 0) {
                pdaArgs.push(pdaSeedsFragment.render);
            }
            return defaultFragment(`await ${pdaFunction}(${pdaArgs.join(', ')})`)
                .mergeImportsWith(pdaSeedsFragment)
                .addImports(pdaImportFrom, pdaFunction);
        case 'publicKeyValueNode':
            return defaultFragment(`'${defaultValue.publicKey}' as Address<'${defaultValue.publicKey}'>`).addImports('solanaAddresses', 'Address');
        case 'programLinkNode':
            const programAddress = nameApi.programAddressConstant(defaultValue.name);
            const importFrom = (_b = defaultValue.importFrom) !== null && _b !== void 0 ? _b : 'generatedPrograms';
            return defaultFragment(programAddress, false).addImports(importFrom, programAddress);
        case 'programIdValueNode':
            if (optionalAccountStrategy === 'programId' &&
                input.kind === 'instructionAccountNode' &&
                input.isOptional) {
                return (0, common_1.fragment)('');
            }
            return defaultFragment('programAddress', false);
        case 'identityValueNode':
        case 'payerValueNode':
            return (0, common_1.fragment)('');
        case 'accountBumpValueNode':
            return defaultFragment(`expectProgramDerivedAddress(accounts.${(0, shared_1.camelCase)(defaultValue.name)}.value)[1]`).addImports('shared', 'expectProgramDerivedAddress');
        case 'argumentValueNode':
            return defaultFragment(`expectSome(args.${(0, shared_1.camelCase)(defaultValue.name)})`).addImports('shared', 'expectSome');
        case 'resolverValueNode':
            const resolverFunction = nameApi.resolverFunction(defaultValue.name);
            const resolverAwait = useAsync && asyncResolvers.includes(defaultValue.name) ? 'await ' : '';
            return defaultFragment(`${resolverAwait}${resolverFunction}(resolverScope)`)
                .addImports((_c = defaultValue.importFrom) !== null && _c !== void 0 ? _c : 'hooked', resolverFunction)
                .addFeatures(['instruction:resolverScopeVariable']);
        case 'conditionalValueNode':
            const ifTrueRenderer = renderNestedInstructionDefault({
                ...scope,
                defaultValue: defaultValue.ifTrue,
            });
            const ifFalseRenderer = renderNestedInstructionDefault({
                ...scope,
                defaultValue: defaultValue.ifFalse,
            });
            if (!ifTrueRenderer && !ifFalseRenderer) {
                return (0, common_1.fragment)('');
            }
            const conditionalFragment = (0, common_1.fragment)('');
            if (ifTrueRenderer) {
                conditionalFragment
                    .mergeImportsWith(ifTrueRenderer)
                    .mergeFeaturesWith(ifTrueRenderer);
            }
            if (ifFalseRenderer) {
                conditionalFragment
                    .mergeImportsWith(ifFalseRenderer)
                    .mergeFeaturesWith(ifFalseRenderer);
            }
            const negatedCondition = !ifTrueRenderer;
            let condition = 'true';
            if ((0, nodes_1.isNode)(defaultValue.condition, 'resolverValueNode')) {
                const conditionalResolverFunction = nameApi.resolverFunction(defaultValue.condition.name);
                conditionalFragment
                    .addImports((_d = defaultValue.condition.importFrom) !== null && _d !== void 0 ? _d : 'hooked', conditionalResolverFunction)
                    .addFeatures(['instruction:resolverScopeVariable']);
                const conditionalResolverAwait = useAsync && asyncResolvers.includes(defaultValue.condition.name)
                    ? 'await '
                    : '';
                condition = `${conditionalResolverAwait}${conditionalResolverFunction}(resolverScope)`;
                condition = negatedCondition ? `!${condition}` : condition;
            }
            else {
                const comparedInputName = (0, nodes_1.isNode)(defaultValue.condition, 'accountValueNode')
                    ? `accounts.${(0, shared_1.camelCase)(defaultValue.condition.name)}.value`
                    : `args.${(0, shared_1.camelCase)(defaultValue.condition.name)}`;
                if (defaultValue.value) {
                    const comparedValue = (0, visitors_1.visit)(defaultValue.value, typeManifestVisitor).value;
                    conditionalFragment
                        .mergeImportsWith(comparedValue)
                        .mergeFeaturesWith(comparedValue);
                    const operator = negatedCondition ? '!==' : '===';
                    condition = `${comparedInputName} ${operator} ${comparedValue.render}`;
                }
                else {
                    condition = negatedCondition
                        ? `!${comparedInputName}`
                        : comparedInputName;
                }
            }
            if (ifTrueRenderer && ifFalseRenderer) {
                return conditionalFragment.setRender(`if (${condition}) {\n${ifTrueRenderer.render}\n} else {\n${ifFalseRenderer.render}\n}`);
            }
            return conditionalFragment.setRender(`if (${condition}) {\n${ifTrueRenderer ? ifTrueRenderer.render : ifFalseRenderer === null || ifFalseRenderer === void 0 ? void 0 : ifFalseRenderer.render}\n}`);
        default:
            const valueManifest = (0, visitors_1.visit)(defaultValue, typeManifestVisitor).value;
            return defaultFragment(valueManifest.render).mergeImportsWith(valueManifest);
    }
}
exports.getInstructionInputDefaultFragment = getInstructionInputDefaultFragment;
function renderNestedInstructionDefault(scope) {
    const { input, defaultValue } = scope;
    if (!defaultValue)
        return undefined;
    return getInstructionInputDefaultFragment({
        ...scope,
        input: { ...input, defaultValue },
    });
}
//# sourceMappingURL=instructionInputDefault.js.map