"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getInstructionInputTypeFragment = void 0;
const nodes_1 = require("../../../nodes");
const shared_1 = require("../../../shared");
const asyncHelpers_1 = require("../asyncHelpers");
const common_1 = require("./common");
function getInstructionInputTypeFragment(scope) {
    const { instructionNode, useAsync, nameApi } = scope;
    const instructionInputType = useAsync
        ? nameApi.instructionAsyncInputType(instructionNode.name)
        : nameApi.instructionSyncInputType(instructionNode.name);
    const accountsFragment = getAccountsFragment(scope);
    const [dataArgumentsFragment, customDataArgumentsFragment] = getDataArgumentsFragments(scope);
    const extraArgumentsFragment = getExtraArgumentsFragment(scope);
    const remainingAccountsFragment = getRemainingAccountsFragment(instructionNode);
    return (0, common_1.fragmentFromTemplate)('instructionInputType.njk', {
        instruction: instructionNode,
        instructionInputType,
        accountsFragment,
        dataArgumentsFragment,
        customDataArgumentsFragment,
        extraArgumentsFragment,
        remainingAccountsFragment,
    })
        .mergeImportsWith(accountsFragment, dataArgumentsFragment, customDataArgumentsFragment, extraArgumentsFragment, remainingAccountsFragment)
        .addImports('solanaAddresses', ['Address']);
}
exports.getInstructionInputTypeFragment = getInstructionInputTypeFragment;
function getAccountsFragment(scope) {
    const { instructionNode, resolvedInputs, useAsync, asyncResolvers } = scope;
    const fragments = instructionNode.accounts.map((account) => {
        const resolvedAccount = resolvedInputs.find((input) => input.kind === 'instructionAccountNode' && input.name === account.name);
        const hasDefaultValue = !!resolvedAccount.defaultValue &&
            !(0, nodes_1.isNode)(resolvedAccount.defaultValue, [
                'identityValueNode',
                'payerValueNode',
            ]) &&
            (useAsync ||
                !(0, asyncHelpers_1.isAsyncDefaultValue)(resolvedAccount.defaultValue, asyncResolvers));
        const docblock = account.docs.length > 0 ? (0, shared_1.jsDocblock)(account.docs) : '';
        const optionalSign = hasDefaultValue || resolvedAccount.isOptional ? '?' : '';
        return getAccountTypeFragment(resolvedAccount).mapRender((r) => `${docblock}${(0, shared_1.camelCase)(account.name)}${optionalSign}: ${r};`);
    });
    return (0, common_1.mergeFragments)(fragments, (r) => r.join('\n'));
}
function getAccountTypeFragment(account) {
    const typeParam = `TAccount${(0, shared_1.pascalCase)(account.name)}`;
    if (account.isPda && account.isSigner === false) {
        return (0, common_1.fragment)(`ProgramDerivedAddress<${typeParam}>`).addImports('solanaAddresses', ['ProgramDerivedAddress']);
    }
    if (account.isPda && account.isSigner === 'either') {
        return (0, common_1.fragment)(`ProgramDerivedAddress<${typeParam}> | TransactionSigner<${typeParam}>`)
            .addImports('solanaAddresses', ['ProgramDerivedAddress'])
            .addImports('solanaSigners', ['TransactionSigner']);
    }
    if (account.isSigner === 'either') {
        return (0, common_1.fragment)(`Address<${typeParam}> | TransactionSigner<${typeParam}>`)
            .addImports('solanaAddresses', ['Address'])
            .addImports('solanaSigners', ['TransactionSigner']);
    }
    if (account.isSigner) {
        return (0, common_1.fragment)(`TransactionSigner<${typeParam}>`).addImports('solanaSigners', ['TransactionSigner']);
    }
    return (0, common_1.fragment)(`Address<${typeParam}>`).addImports('solanaAddresses', [
        'Address',
    ]);
}
function getDataArgumentsFragments(scope) {
    const { instructionNode, nameApi } = scope;
    const customData = scope.customInstructionData.get(instructionNode.name);
    if (customData) {
        return [
            (0, common_1.fragment)(''),
            (0, common_1.fragment)(nameApi.dataArgsType(customData.importAs))
                .mergeImportsWith(scope.dataArgsManifest.looseType)
                .mapRender((r) => `${r} & `),
        ];
    }
    const instructionDataName = nameApi.instructionDataType(instructionNode.name);
    const dataArgsType = nameApi.dataArgsType(instructionDataName);
    const fragments = instructionNode.arguments.flatMap((arg) => {
        const argFragment = getArgumentFragment(arg, (0, common_1.fragment)(dataArgsType), scope.resolvedInputs, scope.renamedArgs);
        return argFragment ? [argFragment] : [];
    });
    return [(0, common_1.mergeFragments)(fragments, (r) => r.join('\n')), (0, common_1.fragment)('')];
}
function getExtraArgumentsFragment(scope) {
    var _a;
    const { instructionNode, nameApi } = scope;
    const instructionExtraName = nameApi.instructionExtraType(instructionNode.name);
    const extraArgsType = nameApi.dataArgsType(instructionExtraName);
    const fragments = ((_a = instructionNode.extraArguments) !== null && _a !== void 0 ? _a : []).flatMap((arg) => {
        const argFragment = getArgumentFragment(arg, (0, common_1.fragment)(extraArgsType), scope.resolvedInputs, scope.renamedArgs);
        return argFragment ? [argFragment] : [];
    });
    return (0, common_1.mergeFragments)(fragments, (r) => r.join('\n'));
}
function getArgumentFragment(arg, argsType, resolvedInputs, renamedArgs) {
    var _a;
    const resolvedArg = resolvedInputs.find((input) => (0, nodes_1.isNode)(input, 'instructionArgumentNode') && input.name === arg.name);
    if (arg.defaultValue && arg.defaultValueStrategy === 'omitted')
        return null;
    const renamedName = (_a = renamedArgs.get(arg.name)) !== null && _a !== void 0 ? _a : arg.name;
    const optionalSign = arg.defaultValue || (resolvedArg === null || resolvedArg === void 0 ? void 0 : resolvedArg.defaultValue) ? '?' : '';
    return argsType.mapRender((r) => `${(0, shared_1.camelCase)(renamedName)}${optionalSign}: ${r}["${(0, shared_1.camelCase)(arg.name)}"];`);
}
function getRemainingAccountsFragment(instructionNode) {
    var _a;
    const fragments = ((_a = instructionNode.remainingAccounts) !== null && _a !== void 0 ? _a : []).flatMap((remainingAccountsNode) => {
        var _a, _b;
        if ((0, nodes_1.isNode)(remainingAccountsNode.value, 'resolverValueNode'))
            return [];
        const { name } = remainingAccountsNode.value;
        const allArguments = (0, nodes_1.getAllInstructionArguments)(instructionNode);
        const argumentExists = allArguments.some((arg) => arg.name === name);
        if (argumentExists)
            return [];
        const isSigner = (_a = remainingAccountsNode.isSigner) !== null && _a !== void 0 ? _a : false;
        const optionalSign = ((_b = remainingAccountsNode.isOptional) !== null && _b !== void 0 ? _b : false) ? '?' : '';
        const signerFragment = (0, common_1.fragment)(`TransactionSigner`).addImports('solanaSigners', ['TransactionSigner']);
        const addressFragment = (0, common_1.fragment)(`Address`).addImports('solanaAddresses', ['Address']);
        return (() => {
            if (isSigner === 'either') {
                return (0, common_1.mergeFragments)([signerFragment, addressFragment], (r) => r.join(' | '));
            }
            return isSigner ? signerFragment : addressFragment;
        })().mapRender((r) => `${(0, shared_1.camelCase)(name)}${optionalSign}: Array<${r}>;`);
    });
    return (0, common_1.mergeFragments)(fragments, (r) => r.join('\n'));
}
//# sourceMappingURL=instructionInputType.js.map