"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getTypeManifestVisitor = void 0;
const nodes_1 = require("../../nodes");
const shared_1 = require("../../shared");
const visitors_1 = require("../../visitors");
const ImportMap_1 = require("./ImportMap");
const TypeManifest_1 = require("./TypeManifest");
const fragments_1 = require("./fragments");
function getTypeManifestVisitor(input) {
    var _a;
    const { nameApi, linkables, nonScalarEnums, customAccountData, customInstructionData, } = input;
    let parentName = (_a = input.parentName) !== null && _a !== void 0 ? _a : null;
    return (0, shared_1.pipe)((0, visitors_1.staticVisitor)(() => ({
        isEnum: false,
        strictType: (0, fragments_1.fragment)(''),
        looseType: (0, fragments_1.fragment)(''),
        encoder: (0, fragments_1.fragment)(''),
        decoder: (0, fragments_1.fragment)(''),
        value: (0, fragments_1.fragment)(''),
    }), [
        ...nodes_1.REGISTERED_TYPE_NODE_KINDS,
        ...nodes_1.REGISTERED_VALUE_NODE_KINDS,
        'definedTypeLinkNode',
        'definedTypeNode',
        'accountNode',
        'instructionNode',
    ]), (visitor) => (0, visitors_1.extendVisitor)(visitor, {
        visitAccount(account, { self }) {
            var _a;
            const accountDataName = nameApi.accountDataType(account.name);
            parentName = {
                strict: nameApi.dataType(accountDataName),
                loose: nameApi.dataArgsType(accountDataName),
            };
            const link = (_a = customAccountData.get(account.name)) === null || _a === void 0 ? void 0 : _a.linkNode;
            const manifest = link ? (0, visitors_1.visit)(link, self) : (0, visitors_1.visit)(account.data, self);
            parentName = null;
            return manifest;
        },
        visitInstruction(instruction, { self }) {
            var _a;
            const instructionDataName = nameApi.instructionDataType(instruction.name);
            parentName = {
                strict: nameApi.dataType(instructionDataName),
                loose: nameApi.dataArgsType(instructionDataName),
            };
            const link = (_a = customInstructionData.get(instruction.name)) === null || _a === void 0 ? void 0 : _a.linkNode;
            const struct = (0, nodes_1.structTypeNodeFromInstructionArgumentNodes)(instruction.arguments);
            const manifest = link ? (0, visitors_1.visit)(link, self) : (0, visitors_1.visit)(struct, self);
            parentName = null;
            return manifest;
        },
        visitDefinedType(definedType, { self }) {
            parentName = {
                strict: nameApi.dataType(definedType.name),
                loose: nameApi.dataArgsType(definedType.name),
            };
            const manifest = (0, visitors_1.visit)(definedType.type, self);
            parentName = null;
            return manifest;
        },
        visitArrayType(arrayType, { self }) {
            const childManifest = (0, visitors_1.visit)(arrayType.item, self);
            childManifest.looseType.mapRender((r) => `Array<${r}>`);
            childManifest.strictType.mapRender((r) => `Array<${r}>`);
            const sizeManifest = getArrayLikeSizeOption(arrayType.count, self);
            const encoderOptions = sizeManifest.encoder.render
                ? `, { ${sizeManifest.encoder.render} }`
                : '';
            const decoderOptions = sizeManifest.decoder.render
                ? `, { ${sizeManifest.decoder.render} }`
                : '';
            childManifest.encoder
                .mapRender((r) => `getArrayEncoder(${r + encoderOptions})`)
                .mergeImportsWith(sizeManifest.encoder)
                .addImports('solanaCodecsDataStructures', 'getArrayEncoder');
            childManifest.decoder
                .mapRender((r) => `getArrayDecoder(${r + decoderOptions})`)
                .mergeImportsWith(sizeManifest.decoder)
                .addImports('solanaCodecsDataStructures', 'getArrayDecoder');
            return childManifest;
        },
        visitDefinedTypeLink(node) {
            var _a;
            const strictName = nameApi.dataType(node.name);
            const looseName = nameApi.dataArgsType(node.name);
            const encoderFunction = nameApi.encoderFunction(node.name);
            const decoderFunction = nameApi.decoderFunction(node.name);
            const importFrom = (_a = node.importFrom) !== null && _a !== void 0 ? _a : 'generatedTypes';
            return {
                isEnum: false,
                strictType: (0, fragments_1.fragment)(strictName).addImports(importFrom, strictName),
                looseType: (0, fragments_1.fragment)(looseName).addImports(importFrom, looseName),
                encoder: (0, fragments_1.fragment)(`${encoderFunction}()`).addImports(importFrom, encoderFunction),
                decoder: (0, fragments_1.fragment)(`${decoderFunction}()`).addImports(importFrom, decoderFunction),
                value: (0, fragments_1.fragment)(''),
            };
        },
        visitEnumType(enumType, { self }) {
            var _a;
            const currentParentName = parentName;
            const encoderImports = new ImportMap_1.ImportMap();
            const decoderImports = new ImportMap_1.ImportMap();
            const encoderOptions = [];
            const decoderOptions = [];
            const enumSize = (0, nodes_1.resolveNestedTypeNode)(enumType.size);
            if (enumSize.format !== 'u8' || enumSize.endian !== 'le') {
                const sizeManifest = (0, visitors_1.visit)(enumType.size, self);
                encoderImports.mergeWith(sizeManifest.encoder);
                decoderImports.mergeWith(sizeManifest.decoder);
                encoderOptions.push(`size: ${sizeManifest.encoder.render}`);
                decoderOptions.push(`size: ${sizeManifest.decoder.render}`);
            }
            const discriminator = nameApi.discriminatedUnionDiscriminator((0, shared_1.mainCase)((_a = currentParentName === null || currentParentName === void 0 ? void 0 : currentParentName.strict) !== null && _a !== void 0 ? _a : ''));
            if (!(0, nodes_1.isScalarEnum)(enumType) && discriminator !== '__kind') {
                encoderOptions.push(`discriminator: '${discriminator}'`);
                decoderOptions.push(`discriminator: '${discriminator}'`);
            }
            const encoderOptionsAsString = encoderOptions.length > 0
                ? `, { ${encoderOptions.join(', ')} }`
                : '';
            const decoderOptionsAsString = decoderOptions.length > 0
                ? `, { ${decoderOptions.join(', ')} }`
                : '';
            if ((0, nodes_1.isScalarEnum)(enumType)) {
                if (currentParentName === null) {
                    throw new Error('Scalar enums cannot be inlined and must be introduced ' +
                        'via a defined type. Ensure you are not inlining a ' +
                        'defined type that is a scalar enum through a visitor.');
                }
                const variantNames = enumType.variants.map(({ name }) => nameApi.enumVariant(name));
                return {
                    isEnum: true,
                    strictType: (0, fragments_1.fragment)(`{ ${variantNames.join(', ')} }`),
                    looseType: (0, fragments_1.fragment)(`{ ${variantNames.join(', ')} }`),
                    encoder: (0, fragments_1.fragment)(`getEnumEncoder(${currentParentName.strict + encoderOptionsAsString})`, encoderImports.add('solanaCodecsDataStructures', 'getEnumEncoder')),
                    decoder: (0, fragments_1.fragment)(`getEnumDecoder(${currentParentName.strict + decoderOptionsAsString})`, decoderImports.add('solanaCodecsDataStructures', 'getEnumDecoder')),
                    value: (0, fragments_1.fragment)(''),
                };
            }
            const mergedManifest = (0, TypeManifest_1.mergeManifests)(enumType.variants.map((variant) => (0, visitors_1.visit)(variant, self)), {
                mergeTypes: (renders) => renders.join(' | '),
                mergeCodecs: (renders) => renders.join(', '),
            });
            mergedManifest.encoder
                .mapRender((r) => `getDiscriminatedUnionEncoder([${r}]${encoderOptionsAsString})`)
                .mergeImportsWith(encoderImports)
                .addImports('solanaCodecsDataStructures', [
                'getDiscriminatedUnionEncoder',
            ]);
            mergedManifest.decoder
                .mapRender((r) => `getDiscriminatedUnionDecoder([${r}]${decoderOptionsAsString})`)
                .mergeImportsWith(decoderImports)
                .addImports('solanaCodecsDataStructures', [
                'getDiscriminatedUnionDecoder',
            ]);
            return mergedManifest;
        },
        visitEnumEmptyVariantType(enumEmptyVariantType) {
            var _a;
            const discriminator = nameApi.discriminatedUnionDiscriminator((0, shared_1.mainCase)((_a = parentName === null || parentName === void 0 ? void 0 : parentName.strict) !== null && _a !== void 0 ? _a : ''));
            const name = nameApi.discriminatedUnionVariant(enumEmptyVariantType.name);
            const kindAttribute = `${discriminator}: "${name}"`;
            return {
                isEnum: false,
                strictType: (0, fragments_1.fragment)(`{ ${kindAttribute} }`),
                looseType: (0, fragments_1.fragment)(`{ ${kindAttribute} }`),
                encoder: (0, fragments_1.fragment)(`['${name}', getUnitEncoder()]`).addImports('solanaCodecsDataStructures', 'getUnitEncoder'),
                decoder: (0, fragments_1.fragment)(`['${name}', getUnitDecoder()]`).addImports('solanaCodecsDataStructures', 'getUnitDecoder'),
                value: (0, fragments_1.fragment)(''),
            };
        },
        visitEnumStructVariantType(enumStructVariantType, { self }) {
            var _a;
            const currentParentName = parentName;
            const discriminator = nameApi.discriminatedUnionDiscriminator((0, shared_1.mainCase)((_a = currentParentName === null || currentParentName === void 0 ? void 0 : currentParentName.strict) !== null && _a !== void 0 ? _a : ''));
            const name = nameApi.discriminatedUnionVariant(enumStructVariantType.name);
            const kindAttribute = `${discriminator}: "${name}"`;
            parentName = null;
            const structManifest = (0, visitors_1.visit)(enumStructVariantType.struct, self);
            parentName = currentParentName;
            structManifest.strictType.mapRender((r) => `{ ${kindAttribute},${r.slice(1, -1)}}`);
            structManifest.looseType.mapRender((r) => `{ ${kindAttribute},${r.slice(1, -1)}}`);
            structManifest.encoder.mapRender((r) => `['${name}', ${r}]`);
            structManifest.decoder.mapRender((r) => `['${name}', ${r}]`);
            return structManifest;
        },
        visitEnumTupleVariantType(enumTupleVariantType, { self }) {
            var _a;
            const currentParentName = parentName;
            const discriminator = nameApi.discriminatedUnionDiscriminator((0, shared_1.mainCase)((_a = currentParentName === null || currentParentName === void 0 ? void 0 : currentParentName.strict) !== null && _a !== void 0 ? _a : ''));
            const name = nameApi.discriminatedUnionVariant(enumTupleVariantType.name);
            const kindAttribute = `${discriminator}: "${name}"`;
            const struct = (0, nodes_1.structTypeNode)([
                (0, nodes_1.structFieldTypeNode)({
                    name: 'fields',
                    type: enumTupleVariantType.tuple,
                }),
            ]);
            parentName = null;
            const structManifest = (0, visitors_1.visit)(struct, self);
            parentName = currentParentName;
            structManifest.strictType.mapRender((r) => `{ ${kindAttribute},${r.slice(1, -1)}}`);
            structManifest.looseType.mapRender((r) => `{ ${kindAttribute},${r.slice(1, -1)}}`);
            structManifest.encoder.mapRender((r) => `['${name}', ${r}]`);
            structManifest.decoder.mapRender((r) => `['${name}', ${r}]`);
            return structManifest;
        },
        visitMapType(mapType, { self }) {
            const key = (0, visitors_1.visit)(mapType.key, self);
            const value = (0, visitors_1.visit)(mapType.value, self);
            const mergedManifest = (0, TypeManifest_1.mergeManifests)([key, value], {
                mergeTypes: ([k, v]) => `Map<${k}, ${v}>`,
                mergeCodecs: ([k, v]) => `${k}, ${v}`,
            });
            const sizeManifest = getArrayLikeSizeOption(mapType.count, self);
            const encoderOptions = sizeManifest.encoder.render
                ? `, { ${sizeManifest.encoder.render} }`
                : '';
            const decoderOptions = sizeManifest.decoder.render
                ? `, { ${sizeManifest.decoder.render} }`
                : '';
            mergedManifest.encoder
                .mapRender((r) => `getMapEncoder(${r}${encoderOptions})`)
                .addImports('solanaCodecsDataStructures', 'getMapEncoder');
            mergedManifest.decoder
                .mapRender((r) => `getMapDecoder(${r}${decoderOptions})`)
                .addImports('solanaCodecsDataStructures', 'getMapDecoder');
            return mergedManifest;
        },
        visitOptionType(optionType, { self }) {
            const childManifest = (0, visitors_1.visit)(optionType.item, self);
            childManifest.strictType
                .mapRender((r) => `Option<${r}>`)
                .addImports('solanaOptions', 'Option');
            childManifest.looseType
                .mapRender((r) => `OptionOrNullable<${r}>`)
                .addImports('solanaOptions', 'OptionOrNullable');
            const encoderOptions = [];
            const decoderOptions = [];
            // Prefix option.
            const optionPrefix = (0, nodes_1.resolveNestedTypeNode)(optionType.prefix);
            if (optionPrefix.format !== 'u8' || optionPrefix.endian !== 'le') {
                const prefixManifest = (0, visitors_1.visit)(optionType.prefix, self);
                childManifest.encoder.mergeImportsWith(prefixManifest.encoder);
                childManifest.decoder.mergeImportsWith(prefixManifest.decoder);
                encoderOptions.push(`prefix: ${prefixManifest.encoder.render}`);
                decoderOptions.push(`prefix: ${prefixManifest.decoder.render}`);
            }
            // Fixed option.
            if (optionType.fixed) {
                encoderOptions.push(`fixed: true`);
                decoderOptions.push(`fixed: true`);
            }
            const encoderOptionsAsString = encoderOptions.length > 0
                ? `, { ${encoderOptions.join(', ')} }`
                : '';
            const decoderOptionsAsString = decoderOptions.length > 0
                ? `, { ${decoderOptions.join(', ')} }`
                : '';
            childManifest.encoder
                .mapRender((r) => `getOptionEncoder(${r + encoderOptionsAsString})`)
                .addImports('solanaOptions', 'getOptionEncoder');
            childManifest.decoder
                .mapRender((r) => `getOptionDecoder(${r + decoderOptionsAsString})`)
                .addImports('solanaOptions', 'getOptionDecoder');
            return childManifest;
        },
        visitSetType(setType, { self }) {
            const childManifest = (0, visitors_1.visit)(setType.item, self);
            childManifest.strictType.mapRender((r) => `Set<${r}>`);
            childManifest.looseType.mapRender((r) => `Set<${r}>`);
            const sizeManifest = getArrayLikeSizeOption(setType.count, self);
            const encoderOptions = sizeManifest.encoder.render
                ? `, { ${sizeManifest.encoder.render} }`
                : '';
            const decoderOptions = sizeManifest.decoder.render
                ? `, { ${sizeManifest.decoder.render} }`
                : '';
            childManifest.encoder
                .mergeImportsWith(sizeManifest.encoder)
                .mapRender((r) => `getSetEncoder(${r + encoderOptions})`)
                .addImports('solanaCodecsDataStructures', 'getSetEncoder');
            childManifest.decoder
                .mergeImportsWith(sizeManifest.decoder)
                .mapRender((r) => `getSetDecoder(${r + decoderOptions})`)
                .addImports('solanaCodecsDataStructures', 'getSetDecoder');
            return childManifest;
        },
        visitStructType(structType, { self }) {
            const optionalFields = structType.fields.filter((f) => !!f.defaultValue);
            const mergedManifest = (0, TypeManifest_1.mergeManifests)(structType.fields.map((field) => (0, visitors_1.visit)(field, self)), {
                mergeTypes: (renders) => `{ ${renders.join('')} }`,
                mergeCodecs: (renders) => `([${renders.join(', ')}])`,
            });
            mergedManifest.encoder
                .mapRender((r) => `getStructEncoder${r}`)
                .addImports('solanaCodecsDataStructures', 'getStructEncoder');
            mergedManifest.decoder
                .mapRender((r) => `getStructDecoder${r}`)
                .addImports('solanaCodecsDataStructures', 'getStructDecoder');
            if (optionalFields.length === 0) {
                return mergedManifest;
            }
            const defaultValues = optionalFields
                .map((f) => {
                const key = (0, shared_1.camelCase)(f.name);
                const defaultValue = f.defaultValue;
                const { render: renderedValue, imports } = (0, visitors_1.visit)(defaultValue, self).value;
                mergedManifest.encoder.mergeImportsWith(imports);
                return f.defaultValueStrategy === 'omitted'
                    ? `${key}: ${renderedValue}`
                    : `${key}: value.${key} ?? ${renderedValue}`;
            })
                .join(', ');
            mergedManifest.encoder
                .mapRender((r) => `transformEncoder(${r}, (value) => ({ ...value, ${defaultValues} }))`)
                .addImports('solanaCodecsCore', 'transformEncoder');
            return mergedManifest;
        },
        visitStructFieldType(structFieldType, { self }) {
            const name = (0, shared_1.camelCase)(structFieldType.name);
            const childManifest = (0, visitors_1.visit)(structFieldType.type, self);
            const docblock = structFieldType.docs.length > 0
                ? `\n${(0, shared_1.jsDocblock)(structFieldType.docs)}`
                : '';
            const originalLooseType = childManifest.looseType.render;
            childManifest.strictType.mapRender((r) => `${docblock}${name}: ${r}; `);
            childManifest.looseType.mapRender((r) => `${docblock}${name}: ${r}; `);
            childManifest.encoder.mapRender((r) => `['${name}', ${r}]`);
            childManifest.decoder.mapRender((r) => `['${name}', ${r}]`);
            // No default value.
            if (!structFieldType.defaultValue) {
                return childManifest;
            }
            // Optional default value.
            if (structFieldType.defaultValueStrategy !== 'omitted') {
                childManifest.looseType.setRender(`${docblock}${name}?: ${originalLooseType}; `);
                return childManifest;
            }
            // Omitted default value.
            childManifest.looseType = (0, fragments_1.fragment)('');
            return childManifest;
        },
        visitTupleType(tupleType, { self }) {
            const items = tupleType.items.map((item) => (0, visitors_1.visit)(item, self));
            const mergedManifest = (0, TypeManifest_1.mergeManifests)(items, {
                mergeTypes: (types) => `readonly [${types.join(', ')}]`,
                mergeCodecs: (codecs) => `[${codecs.join(', ')}]`,
            });
            mergedManifest.encoder
                .mapRender((render) => `getTupleEncoder(${render})`)
                .addImports('solanaCodecsDataStructures', 'getTupleEncoder');
            mergedManifest.decoder
                .mapRender((render) => `getTupleDecoder(${render})`)
                .addImports('solanaCodecsDataStructures', 'getTupleDecoder');
            return mergedManifest;
        },
        visitBooleanType(booleanType, { self }) {
            const encoderImports = new ImportMap_1.ImportMap().add('solanaCodecsDataStructures', 'getBooleanEncoder');
            const decoderImports = new ImportMap_1.ImportMap().add('solanaCodecsDataStructures', 'getBooleanDecoder');
            let sizeEncoder = '';
            let sizeDecoder = '';
            const resolvedSize = (0, nodes_1.resolveNestedTypeNode)(booleanType.size);
            if (resolvedSize.format !== 'u8' || resolvedSize.endian !== 'le') {
                const size = (0, visitors_1.visit)(booleanType.size, self);
                encoderImports.mergeWith(size.encoder);
                decoderImports.mergeWith(size.decoder);
                sizeEncoder = `{ size: ${size.encoder.render} }`;
                sizeDecoder = `{ size: ${size.decoder.render} }`;
            }
            return {
                isEnum: false,
                strictType: (0, fragments_1.fragment)('boolean'),
                looseType: (0, fragments_1.fragment)('boolean'),
                encoder: (0, fragments_1.fragment)(`getBooleanEncoder(${sizeEncoder})`, encoderImports),
                decoder: (0, fragments_1.fragment)(`getBooleanDecoder(${sizeDecoder})`, decoderImports),
                value: (0, fragments_1.fragment)(''),
            };
        },
        visitBytesType() {
            return {
                isEnum: false,
                strictType: (0, fragments_1.fragment)('ReadonlyUint8Array').addImports('solanaCodecsCore', 'ReadonlyUint8Array'),
                looseType: (0, fragments_1.fragment)('ReadonlyUint8Array').addImports('solanaCodecsCore', 'ReadonlyUint8Array'),
                encoder: (0, fragments_1.fragment)(`getBytesEncoder()`).addImports('solanaCodecsDataStructures', 'getBytesEncoder'),
                decoder: (0, fragments_1.fragment)(`getBytesDecoder()`).addImports('solanaCodecsDataStructures', 'getBytesDecoder'),
                value: (0, fragments_1.fragment)(''),
            };
        },
        visitNumberType(numberType) {
            const encoderFunction = nameApi.encoderFunction(numberType.format);
            const decoderFunction = nameApi.decoderFunction(numberType.format);
            const isBigNumber = ['u64', 'u128', 'i64', 'i128'].includes(numberType.format);
            const encoderImports = new ImportMap_1.ImportMap().add('solanaCodecsNumbers', encoderFunction);
            const decoderImports = new ImportMap_1.ImportMap().add('solanaCodecsNumbers', decoderFunction);
            let endianness = '';
            if (numberType.endian === 'be') {
                encoderImports.add('solanaCodecsNumbers', 'Endian');
                decoderImports.add('solanaCodecsNumbers', 'Endian');
                endianness = '{ endian: Endian.Big }';
            }
            return {
                isEnum: false,
                strictType: (0, fragments_1.fragment)(isBigNumber ? 'bigint' : 'number'),
                looseType: (0, fragments_1.fragment)(isBigNumber ? 'number | bigint' : 'number'),
                encoder: (0, fragments_1.fragment)(`${encoderFunction}(${endianness})`, encoderImports),
                decoder: (0, fragments_1.fragment)(`${decoderFunction}(${endianness})`, decoderImports),
                value: (0, fragments_1.fragment)(''),
            };
        },
        visitAmountType(amountType, { self }) {
            return (0, visitors_1.visit)(amountType.number, self);
        },
        visitDateTimeType(dateTimeType, { self }) {
            return (0, visitors_1.visit)(dateTimeType.number, self);
        },
        visitSolAmountType(solAmountType, { self }) {
            return (0, visitors_1.visit)(solAmountType.number, self);
        },
        visitPublicKeyType() {
            const imports = new ImportMap_1.ImportMap().add('solanaAddresses', 'Address');
            return {
                isEnum: false,
                strictType: (0, fragments_1.fragment)('Address', imports),
                looseType: (0, fragments_1.fragment)('Address', imports),
                encoder: (0, fragments_1.fragment)('getAddressEncoder()').addImports('solanaAddresses', 'getAddressEncoder'),
                decoder: (0, fragments_1.fragment)('getAddressDecoder()').addImports('solanaAddresses', 'getAddressDecoder'),
                value: (0, fragments_1.fragment)(''),
            };
        },
        visitStringType(stringType) {
            const [encoder, decoder] = (() => {
                switch (stringType.encoding) {
                    case 'base16':
                        return ['getBase16Encoder', 'getBase16Decoder'];
                    case 'base58':
                        return ['getBase58Encoder', 'getBase58Decoder'];
                    case 'base64':
                        return ['getBase64Encoder', 'getBase64Decoder'];
                    case 'utf8':
                        return ['getUtf8Encoder', 'getUtf8Decoder'];
                    default:
                        throw new Error(`Unsupported string encoding: ${stringType.encoding}`);
                }
            })();
            return {
                isEnum: false,
                strictType: (0, fragments_1.fragment)('string'),
                looseType: (0, fragments_1.fragment)('string'),
                encoder: (0, fragments_1.fragment)(`${encoder}()`).addImports('solanaCodecsStrings', encoder),
                decoder: (0, fragments_1.fragment)(`${decoder}()`).addImports('solanaCodecsStrings', decoder),
                value: (0, fragments_1.fragment)(''),
            };
        },
        visitFixedSizeType(node, { self }) {
            const manifest = (0, visitors_1.visit)(node.type, self);
            manifest.encoder
                .mapRender((r) => `fixEncoderSize(${r}, ${node.size})`)
                .addImports('solanaCodecsCore', 'fixEncoderSize');
            manifest.decoder
                .mapRender((r) => `fixDecoderSize(${r}, ${node.size})`)
                .addImports('solanaCodecsCore', 'fixDecoderSize');
            return manifest;
        },
        visitHiddenPrefixType(node, { self }) {
            const manifest = (0, visitors_1.visit)(node.type, self);
            const prefixes = node.prefix.map((c) => (0, visitors_1.visit)(c, self).value);
            const prefixEncoders = (0, fragments_1.fragment)(prefixes.map((c) => `getConstantEncoder(${c})`).join(', '))
                .addImports('solanaCodecsCore', 'getConstantEncoder')
                .mergeImportsWith(...prefixes);
            const prefixDecoders = (0, fragments_1.fragment)(prefixes.map((c) => `getConstantDecoder(${c})`).join(', '))
                .addImports('solanaCodecsCore', 'getConstantDecoder')
                .mergeImportsWith(...prefixes);
            manifest.encoder
                .mapRender((r) => `getHiddenPrefixEncoder(${r}, [${prefixEncoders}])`)
                .mergeImportsWith(prefixEncoders)
                .addImports('solanaCodecsDataStructures', 'getHiddenPrefixEncoder');
            manifest.decoder
                .mapRender((r) => `getHiddenPrefixDecoder(${r}, [${prefixDecoders}])`)
                .mergeImportsWith(prefixDecoders)
                .addImports('solanaCodecsDataStructures', 'getHiddenPrefixDecoder');
            return manifest;
        },
        visitHiddenSuffixType(node, { self }) {
            const manifest = (0, visitors_1.visit)(node.type, self);
            const suffixes = node.suffix.map((c) => (0, visitors_1.visit)(c, self).value);
            const suffixEncoders = (0, fragments_1.fragment)(suffixes.map((c) => `getConstantEncoder(${c})`).join(', '))
                .addImports('solanaCodecsCore', 'getConstantEncoder')
                .mergeImportsWith(...suffixes);
            const suffixDecoders = (0, fragments_1.fragment)(suffixes.map((c) => `getConstantDecoder(${c})`).join(', '))
                .addImports('solanaCodecsCore', 'getConstantDecoder')
                .mergeImportsWith(...suffixes);
            manifest.encoder
                .mapRender((r) => `getHiddenSuffixEncoder(${r}, [${suffixEncoders}])`)
                .mergeImportsWith(suffixEncoders)
                .addImports('solanaCodecsDataStructures', 'getHiddenSuffixEncoder');
            manifest.decoder
                .mapRender((r) => `getHiddenSuffixDecoder(${r}, [${suffixDecoders}])`)
                .mergeImportsWith(suffixDecoders)
                .addImports('solanaCodecsDataStructures', 'getHiddenSuffixDecoder');
            return manifest;
        },
        visitPostOffsetType(node, { self }) {
            const manifest = (0, visitors_1.visit)(node.type, self);
            if (node.strategy === 'padded') {
                manifest.encoder
                    .mapRender((r) => `padRightEncoder(${r}, ${node.offset})`)
                    .addImports('solanaCodecsCore', 'padRightEncoder');
                manifest.decoder
                    .mapRender((r) => `padRightDecoder(${r}, ${node.offset})`)
                    .addImports('solanaCodecsCore', 'padRightDecoder');
                return manifest;
            }
            const fn = (() => {
                switch (node.strategy) {
                    case 'absolute':
                        return node.offset < 0
                            ? `({ wrapBytes }) => wrapBytes(${node.offset})`
                            : `() => ${node.offset}`;
                    case 'preOffset':
                        return node.offset < 0
                            ? `({ preOffset }) => preOffset ${node.offset}`
                            : `({ preOffset }) => preOffset + ${node.offset}`;
                    case 'relative':
                    default:
                        return node.offset < 0
                            ? `({ postOffset }) => postOffset ${node.offset}`
                            : `({ postOffset }) => postOffset + ${node.offset}`;
                }
            })();
            manifest.encoder
                .mapRender((r) => `offsetEncoder(${r}, { postOffset: ${fn} })`)
                .addImports('solanaCodecsCore', 'offsetEncoder');
            manifest.decoder
                .mapRender((r) => `offsetDecoder(${r}, { postOffset: ${fn} })`)
                .addImports('solanaCodecsCore', 'offsetDecoder');
            return manifest;
        },
        visitPreOffsetType(node, { self }) {
            const manifest = (0, visitors_1.visit)(node.type, self);
            if (node.strategy === 'padded') {
                manifest.encoder
                    .mapRender((r) => `padLeftEncoder(${r}, ${node.offset})`)
                    .addImports('solanaCodecsCore', 'padLeftEncoder');
                manifest.decoder
                    .mapRender((r) => `padLeftDecoder(${r}, ${node.offset})`)
                    .addImports('solanaCodecsCore', 'padLeftDecoder');
                return manifest;
            }
            const fn = (() => {
                switch (node.strategy) {
                    case 'absolute':
                        return node.offset < 0
                            ? `({ wrapBytes }) => wrapBytes(${node.offset})`
                            : `() => ${node.offset}`;
                    case 'relative':
                    default:
                        return node.offset < 0
                            ? `({ preOffset }) => preOffset ${node.offset}`
                            : `({ preOffset }) => preOffset + ${node.offset}`;
                }
            })();
            manifest.encoder
                .mapRender((r) => `offsetEncoder(${r}, { preOffset: ${fn} })`)
                .addImports('solanaCodecsCore', 'offsetEncoder');
            manifest.decoder
                .mapRender((r) => `offsetDecoder(${r}, { preOffset: ${fn} })`)
                .addImports('solanaCodecsCore', 'offsetDecoder');
            return manifest;
        },
        visitSentinelType(node, { self }) {
            const manifest = (0, visitors_1.visit)(node.type, self);
            const sentinel = (0, visitors_1.visit)(node.sentinel, self).value;
            manifest.encoder
                .mapRender((r) => `addEncoderSentinel(${r}, ${sentinel})`)
                .mergeImportsWith(sentinel)
                .addImports('solanaCodecsCore', 'addEncoderSentinel');
            manifest.decoder
                .mapRender((r) => `addDecoderSentinel(${r}, ${sentinel})`)
                .mergeImportsWith(sentinel)
                .addImports('solanaCodecsCore', 'addDecoderSentinel');
            return manifest;
        },
        visitSizePrefixType(node, { self }) {
            const manifest = (0, visitors_1.visit)(node.type, self);
            const prefix = (0, visitors_1.visit)(node.prefix, self);
            manifest.encoder
                .mapRender((r) => `addEncoderSizePrefix(${r}, ${prefix.encoder})`)
                .mergeImportsWith(prefix.encoder)
                .addImports('solanaCodecsCore', 'addEncoderSizePrefix');
            manifest.decoder
                .mapRender((r) => `addDecoderSizePrefix(${r}, ${prefix.decoder})`)
                .mergeImportsWith(prefix.decoder)
                .addImports('solanaCodecsCore', 'addDecoderSizePrefix');
            return manifest;
        },
        visitArrayValue(node, { self }) {
            return (0, TypeManifest_1.mergeManifests)(node.items.map((v) => (0, visitors_1.visit)(v, self)), { mergeValues: (renders) => `[${renders.join(', ')}]` });
        },
        visitBooleanValue(node) {
            const manifest = (0, TypeManifest_1.typeManifest)();
            manifest.value.setRender(JSON.stringify(node.boolean));
            return manifest;
        },
        visitBytesValue(node) {
            const manifest = (0, TypeManifest_1.typeManifest)();
            const bytes = (0, nodes_1.getBytesFromBytesValueNode)(node);
            manifest.value.setRender(`new Uint8Array([${Array.from(bytes).join(', ')}])`);
            return manifest;
        },
        visitConstantValue(node, { self }) {
            if ((0, nodes_1.isNode)(node.type, 'bytesTypeNode') &&
                (0, nodes_1.isNode)(node.value, 'bytesValueNode')) {
                return (0, visitors_1.visit)(node.value, self);
            }
            return {
                ...(0, TypeManifest_1.typeManifest)(),
                value: (0, fragments_1.mergeFragments)([(0, visitors_1.visit)(node.type, self).encoder, (0, visitors_1.visit)(node.value, self).value], ([encoderFunction, value]) => `${encoderFunction}.encode(${value})`),
            };
        },
        visitEnumValue(node, { self }) {
            var _a, _b;
            const manifest = (0, TypeManifest_1.typeManifest)();
            const enumName = nameApi.dataType(node.enum.name);
            const enumFunction = nameApi.discriminatedUnionFunction(node.enum.name);
            const importFrom = (_a = node.enum.importFrom) !== null && _a !== void 0 ? _a : 'generatedTypes';
            const enumNode = (_b = linkables.get(node.enum)) === null || _b === void 0 ? void 0 : _b.type;
            const isScalar = enumNode && (0, nodes_1.isNode)(enumNode, 'enumTypeNode')
                ? (0, nodes_1.isScalarEnum)(enumNode)
                : !nonScalarEnums.includes(node.enum.name);
            if (!node.value && isScalar) {
                const variantName = nameApi.enumVariant(node.variant);
                manifest.value
                    .setRender(`${enumName}.${variantName}`)
                    .addImports(importFrom, enumName);
                return manifest;
            }
            const variantName = nameApi.discriminatedUnionVariant(node.variant);
            if (!node.value) {
                manifest.value
                    .setRender(`${enumFunction}('${variantName}')`)
                    .addImports(importFrom, enumFunction);
                return manifest;
            }
            manifest.value = (0, visitors_1.visit)(node.value, self)
                .value.mapRender((r) => `${enumFunction}('${variantName}', ${r})`)
                .addImports(importFrom, enumFunction);
            return manifest;
        },
        visitMapValue(node, { self }) {
            const entryFragments = node.entries.map((entry) => (0, visitors_1.visit)(entry, self));
            return (0, TypeManifest_1.mergeManifests)(entryFragments, {
                mergeValues: (renders) => `new Map([${renders.join(', ')}])`,
            });
        },
        visitMapEntryValue(node, { self }) {
            return (0, TypeManifest_1.mergeManifests)([(0, visitors_1.visit)(node.key, self), (0, visitors_1.visit)(node.value, self)], { mergeValues: (renders) => `[${renders.join(', ')}]` });
        },
        visitNoneValue() {
            const manifest = (0, TypeManifest_1.typeManifest)();
            manifest.value
                .setRender('none()')
                .addImports('solanaOptions', 'none');
            return manifest;
        },
        visitNumberValue(node) {
            const manifest = (0, TypeManifest_1.typeManifest)();
            manifest.value.setRender(JSON.stringify(node.number));
            return manifest;
        },
        visitPublicKeyValue(node) {
            const manifest = (0, TypeManifest_1.typeManifest)();
            manifest.value
                .setRender(`address("${node.publicKey}")`)
                .addImports('solanaAddresses', 'address');
            return manifest;
        },
        visitSetValue(node, { self }) {
            return (0, TypeManifest_1.mergeManifests)(node.items.map((v) => (0, visitors_1.visit)(v, self)), { mergeValues: (renders) => `new Set([${renders.join(', ')}])` });
        },
        visitSomeValue(node, { self }) {
            const manifest = (0, TypeManifest_1.typeManifest)();
            manifest.value = (0, visitors_1.visit)(node.value, self)
                .value.mapRender((r) => `some(${r})`)
                .addImports('solanaOptions', 'some');
            return manifest;
        },
        visitStringValue(node) {
            const manifest = (0, TypeManifest_1.typeManifest)();
            manifest.value.setRender(JSON.stringify(node.string));
            return manifest;
        },
        visitStructValue(node, { self }) {
            return (0, TypeManifest_1.mergeManifests)(node.fields.map((field) => (0, visitors_1.visit)(field, self)), { mergeValues: (renders) => `{ ${renders.join(', ')} }` });
        },
        visitStructFieldValue(node, { self }) {
            const manifest = (0, TypeManifest_1.typeManifest)();
            manifest.value = (0, visitors_1.visit)(node.value, self).value.mapRender((r) => `${node.name}: ${r}`);
            return manifest;
        },
        visitTupleValue(node, { self }) {
            return (0, TypeManifest_1.mergeManifests)(node.items.map((v) => (0, visitors_1.visit)(v, self)), { mergeValues: (renders) => `[${renders.join(', ')}]` });
        },
    }));
}
exports.getTypeManifestVisitor = getTypeManifestVisitor;
function getArrayLikeSizeOption(count, visitor) {
    if ((0, nodes_1.isNode)(count, 'fixedCountNode')) {
        return {
            encoder: (0, fragments_1.fragment)(`size: ${count.value}`),
            decoder: (0, fragments_1.fragment)(`size: ${count.value}`),
        };
    }
    if ((0, nodes_1.isNode)(count, 'remainderCountNode')) {
        return {
            encoder: (0, fragments_1.fragment)(`size: 'remainder'`),
            decoder: (0, fragments_1.fragment)(`size: 'remainder'`),
        };
    }
    const prefix = (0, nodes_1.resolveNestedTypeNode)(count.prefix);
    if (prefix.format === 'u32' && prefix.endian === 'le') {
        return { encoder: (0, fragments_1.fragment)(''), decoder: (0, fragments_1.fragment)('') };
    }
    const prefixManifest = (0, visitors_1.visit)(count.prefix, visitor);
    prefixManifest.encoder.mapRender((r) => `size: ${r}`);
    prefixManifest.decoder.mapRender((r) => `size: ${r}`);
    return prefixManifest;
}
//# sourceMappingURL=getTypeManifestVisitor.js.map