"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.RustImportMap = void 0;
const DEFAULT_MODULE_MAP = {
    generated: 'crate::generated',
    generatedAccounts: 'crate::generated::accounts',
    generatedTypes: 'crate::generated::types',
    generatedErrors: 'crate::generated::errors',
    hooked: 'crate::hooked',
    mplEssentials: 'mpl_toolbox',
    mplToolbox: 'mpl_toolbox',
};
class RustImportMap {
    constructor() {
        this._imports = new Set();
        this._aliases = new Map();
    }
    get imports() {
        return this._imports;
    }
    get aliases() {
        return this._aliases;
    }
    add(imports) {
        const newImports = typeof imports === 'string' ? [imports] : imports;
        newImports.forEach((i) => this._imports.add(i));
        return this;
    }
    remove(imports) {
        const importsToRemove = typeof imports === 'string' ? [imports] : imports;
        importsToRemove.forEach((i) => this._imports.delete(i));
        return this;
    }
    mergeWith(...others) {
        others.forEach((other) => {
            this.add(other._imports);
            other._aliases.forEach((alias, importName) => this.addAlias(importName, alias));
        });
        return this;
    }
    mergeWithManifest(manifest) {
        return this.mergeWith(manifest.imports);
    }
    addAlias(importName, alias) {
        this._aliases.set(importName, alias);
        return this;
    }
    isEmpty() {
        return this._imports.size === 0;
    }
    resolveDependencyMap(dependencies) {
        const dependencyMap = { ...DEFAULT_MODULE_MAP, ...dependencies };
        const newImportMap = new RustImportMap();
        const resolveDependency = (i) => {
            const dependencyKey = Object.keys(dependencyMap).find((key) => i.startsWith(`${key}::`));
            if (!dependencyKey)
                return i;
            const dependencyValue = dependencyMap[dependencyKey];
            return dependencyValue + i.slice(dependencyKey.length);
        };
        this._imports.forEach((i) => newImportMap.add(resolveDependency(i)));
        this._aliases.forEach((alias, i) => newImportMap.addAlias(resolveDependency(i), alias));
        return newImportMap;
    }
    toString(dependencies) {
        const resolvedMap = this.resolveDependencyMap(dependencies);
        const importStatements = [...resolvedMap.imports].map((i) => {
            const alias = resolvedMap.aliases.get(i);
            if (alias)
                return `use ${i} as ${alias};`;
            return `use ${i};`;
        });
        return importStatements.join('\n');
    }
}
exports.RustImportMap = RustImportMap;
//# sourceMappingURL=RustImportMap.js.map