"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getTypeManifestVisitor = void 0;
const nodes_1 = require("../../nodes");
const shared_1 = require("../../shared");
const visitors_1 = require("../../visitors");
const RustImportMap_1 = require("./RustImportMap");
function getTypeManifestVisitor() {
    let parentName = null;
    let nestedStruct = false;
    let inlineStruct = false;
    let parentSize = null;
    const visitor = (0, shared_1.pipe)((0, visitors_1.mergeVisitor)(() => ({ type: '', imports: new RustImportMap_1.RustImportMap(), nestedStructs: [] }), (_, values) => ({
        ...mergeManifests(values),
        type: values.map((v) => v.type).join('\n'),
    }), [
        ...nodes_1.REGISTERED_TYPE_NODE_KINDS,
        'definedTypeLinkNode',
        'definedTypeNode',
        'accountNode',
    ]), (v) => (0, visitors_1.extendVisitor)(v, {
        visitAccount(account, { self }) {
            parentName = (0, shared_1.pascalCase)(account.name);
            const manifest = (0, visitors_1.visit)(account.data, self);
            manifest.imports.add([
                'borsh::BorshSerialize',
                'borsh::BorshDeserialize',
            ]);
            parentName = null;
            return {
                ...manifest,
                type: '#[derive(BorshSerialize, BorshDeserialize, Clone, Debug, Eq, PartialEq)]\n' +
                    '#[cfg_attr(feature = "serde", derive(serde::Serialize, serde::Deserialize))]\n' +
                    `${manifest.type}`,
            };
        },
        visitDefinedType(definedType, { self }) {
            parentName = (0, shared_1.pascalCase)(definedType.name);
            const manifest = (0, visitors_1.visit)(definedType.type, self);
            parentName = null;
            manifest.imports.add([
                'borsh::BorshSerialize',
                'borsh::BorshDeserialize',
            ]);
            const traits = [
                'BorshSerialize',
                'BorshDeserialize',
                'Clone',
                'Debug',
                'Eq',
                'PartialEq',
            ];
            if ((0, nodes_1.isNode)(definedType.type, 'enumTypeNode') &&
                (0, nodes_1.isScalarEnum)(definedType.type)) {
                traits.push('PartialOrd', 'Hash', 'FromPrimitive');
                manifest.imports.add(['num_derive::FromPrimitive']);
            }
            return {
                ...manifest,
                type: `#[derive(${traits.join(', ')})]\n` +
                    '#[cfg_attr(feature = "serde", derive(serde::Serialize, serde::Deserialize))]\n' +
                    `${manifest.type}`,
                nestedStructs: manifest.nestedStructs.map((struct) => `#[derive(${traits.join(', ')})]\n` +
                    '#[cfg_attr(feature = "serde", derive(serde::Serialize, serde::Deserialize))]\n' +
                    `${struct}`),
            };
        },
        visitArrayType(arrayType, { self }) {
            const childManifest = (0, visitors_1.visit)(arrayType.item, self);
            if ((0, nodes_1.isNode)(arrayType.count, 'fixedCountNode')) {
                return {
                    ...childManifest,
                    type: `[${childManifest.type}; ${arrayType.count.value}]`,
                };
            }
            if ((0, nodes_1.isNode)(arrayType.count, 'remainderCountNode')) {
                childManifest.imports.add('kaigan::types::RemainderVec');
                return {
                    ...childManifest,
                    type: `RemainderVec<${childManifest.type}>`,
                };
            }
            const prefix = (0, nodes_1.resolveNestedTypeNode)(arrayType.count.prefix);
            if (prefix.endian === 'le') {
                switch (prefix.format) {
                    case 'u32':
                        return {
                            ...childManifest,
                            type: `Vec<${childManifest.type}>`,
                        };
                    case 'u8':
                    case 'u16':
                    case 'u64': {
                        const prefixFormat = prefix.format.toUpperCase();
                        childManifest.imports.add(`kaigan::types::${prefixFormat}PrefixVec`);
                        return {
                            ...childManifest,
                            type: `${prefixFormat}PrefixVec<${childManifest.type}>`,
                        };
                    }
                    default:
                        throw new Error(`Array prefix not supported: ${prefix.format}`);
                }
            }
            // TODO: Add to the Rust validator.
            throw new Error('Array size not supported by Borsh');
        },
        visitDefinedTypeLink(node) {
            var _a;
            const pascalCaseDefinedType = (0, shared_1.pascalCase)(node.name);
            const importFrom = (_a = node.importFrom) !== null && _a !== void 0 ? _a : 'generatedTypes';
            return {
                imports: new RustImportMap_1.RustImportMap().add(`${importFrom}::${pascalCaseDefinedType}`),
                type: pascalCaseDefinedType,
                nestedStructs: [],
            };
        },
        visitEnumType(enumType, { self }) {
            const originalParentName = parentName;
            if (!originalParentName) {
                // TODO: Add to the Rust validator.
                throw new Error('Enum type must have a parent name.');
            }
            const variants = enumType.variants.map((variant) => (0, visitors_1.visit)(variant, self));
            const variantNames = variants
                .map((variant) => variant.type)
                .join('\n');
            const mergedManifest = mergeManifests(variants);
            return {
                ...mergedManifest,
                type: `pub enum ${(0, shared_1.pascalCase)(originalParentName)} {\n${variantNames}\n}`,
            };
        },
        visitEnumEmptyVariantType(enumEmptyVariantType) {
            const name = (0, shared_1.pascalCase)(enumEmptyVariantType.name);
            return {
                type: `${name},`,
                imports: new RustImportMap_1.RustImportMap(),
                nestedStructs: [],
            };
        },
        visitEnumStructVariantType(enumStructVariantType, { self }) {
            const name = (0, shared_1.pascalCase)(enumStructVariantType.name);
            const originalParentName = parentName;
            if (!originalParentName) {
                throw new Error('Enum struct variant type must have a parent name.');
            }
            inlineStruct = true;
            parentName = (0, shared_1.pascalCase)(originalParentName) + name;
            const typeManifest = (0, visitors_1.visit)(enumStructVariantType.struct, self);
            inlineStruct = false;
            parentName = originalParentName;
            return {
                ...typeManifest,
                type: `${name} ${typeManifest.type},`,
            };
        },
        visitEnumTupleVariantType(enumTupleVariantType, { self }) {
            const name = (0, shared_1.pascalCase)(enumTupleVariantType.name);
            const originalParentName = parentName;
            if (!originalParentName) {
                throw new Error('Enum struct variant type must have a parent name.');
            }
            parentName = (0, shared_1.pascalCase)(originalParentName) + name;
            const childManifest = (0, visitors_1.visit)(enumTupleVariantType.tuple, self);
            parentName = originalParentName;
            return {
                ...childManifest,
                type: `${name}${childManifest.type},`,
            };
        },
        visitMapType(mapType, { self }) {
            const key = (0, visitors_1.visit)(mapType.key, self);
            const value = (0, visitors_1.visit)(mapType.value, self);
            const mergedManifest = mergeManifests([key, value]);
            mergedManifest.imports.add('std::collections::HashMap');
            return {
                ...mergedManifest,
                type: `HashMap<${key.type}, ${value.type}>`,
            };
        },
        visitOptionType(optionType, { self }) {
            const childManifest = (0, visitors_1.visit)(optionType.item, self);
            const optionPrefix = (0, nodes_1.resolveNestedTypeNode)(optionType.prefix);
            if (optionPrefix.format === 'u8' && optionPrefix.endian === 'le') {
                return {
                    ...childManifest,
                    type: `Option<${childManifest.type}>`,
                };
            }
            // TODO: Add to the Rust validator.
            throw new Error('Option size not supported by Borsh');
        },
        visitSetType(setType, { self }) {
            const childManifest = (0, visitors_1.visit)(setType.item, self);
            childManifest.imports.add('std::collections::HashSet');
            return {
                ...childManifest,
                type: `HashSet<${childManifest.type}>`,
            };
        },
        visitStructType(structType, { self }) {
            const originalParentName = parentName;
            if (!originalParentName) {
                // TODO: Add to the Rust validator.
                throw new Error('Struct type must have a parent name.');
            }
            const fields = structType.fields.map((field) => (0, visitors_1.visit)(field, self));
            const fieldTypes = fields.map((field) => field.type).join('\n');
            const mergedManifest = mergeManifests(fields);
            if (nestedStruct) {
                return {
                    ...mergedManifest,
                    type: (0, shared_1.pascalCase)(originalParentName),
                    nestedStructs: [
                        ...mergedManifest.nestedStructs,
                        `pub struct ${(0, shared_1.pascalCase)(originalParentName)} {\n${fieldTypes}\n}`,
                    ],
                };
            }
            if (inlineStruct) {
                return { ...mergedManifest, type: `{\n${fieldTypes}\n}` };
            }
            return {
                ...mergedManifest,
                type: `pub struct ${(0, shared_1.pascalCase)(originalParentName)} {\n${fieldTypes}\n}`,
            };
        },
        visitStructFieldType(structFieldType, { self }) {
            const originalParentName = parentName;
            const originalInlineStruct = inlineStruct;
            const originalNestedStruct = nestedStruct;
            if (!originalParentName) {
                throw new Error('Struct field type must have a parent name.');
            }
            parentName =
                (0, shared_1.pascalCase)(originalParentName) + (0, shared_1.pascalCase)(structFieldType.name);
            nestedStruct = true;
            inlineStruct = false;
            const fieldManifest = (0, visitors_1.visit)(structFieldType.type, self);
            parentName = originalParentName;
            inlineStruct = originalInlineStruct;
            nestedStruct = originalNestedStruct;
            const fieldName = (0, shared_1.snakeCase)(structFieldType.name);
            const docblock = (0, shared_1.rustDocblock)(structFieldType.docs);
            const resolvedNestedType = (0, nodes_1.resolveNestedTypeNode)(structFieldType.type);
            let derive = '';
            if (fieldManifest.type === 'Pubkey') {
                derive =
                    '#[cfg_attr(feature = "serde", serde(with = "serde_with::As::<serde_with::DisplayFromStr>"))]\n';
            }
            else if (((0, nodes_1.isNode)(resolvedNestedType, 'arrayTypeNode') &&
                (0, nodes_1.isNode)(resolvedNestedType.count, 'fixedCountNode') &&
                resolvedNestedType.count.value > 32) ||
                ((0, nodes_1.isNode)(resolvedNestedType, ['bytesTypeNode', 'stringTypeNode']) &&
                    (0, nodes_1.isNode)(structFieldType.type, 'fixedSizeTypeNode') &&
                    structFieldType.type.size > 32)) {
                derive =
                    '#[cfg_attr(feature = "serde", serde(with = "serde_with::As::<serde_with::Bytes>"))]\n';
            }
            return {
                ...fieldManifest,
                type: inlineStruct
                    ? `${docblock}${derive}${fieldName}: ${fieldManifest.type},`
                    : `${docblock}${derive}pub ${fieldName}: ${fieldManifest.type},`,
            };
        },
        visitTupleType(tupleType, { self }) {
            const items = tupleType.items.map((item) => (0, visitors_1.visit)(item, self));
            const mergedManifest = mergeManifests(items);
            return {
                ...mergedManifest,
                type: `(${items.map((item) => item.type).join(', ')})`,
            };
        },
        visitBooleanType(booleanType) {
            const resolvedSize = (0, nodes_1.resolveNestedTypeNode)(booleanType.size);
            if (resolvedSize.format === 'u8' && resolvedSize.endian === 'le') {
                return {
                    type: 'bool',
                    imports: new RustImportMap_1.RustImportMap(),
                    nestedStructs: [],
                };
            }
            // TODO: Add to the Rust validator.
            throw new Error('Bool size not supported by Borsh');
        },
        visitBytesType(_bytesType, { self }) {
            let arraySize = (0, nodes_1.remainderCountNode)();
            if (typeof parentSize === 'number') {
                arraySize = (0, nodes_1.fixedCountNode)(parentSize);
            }
            else if (parentSize && typeof parentSize === 'object') {
                arraySize = (0, nodes_1.prefixedCountNode)(parentSize);
            }
            const arrayType = (0, nodes_1.arrayTypeNode)((0, nodes_1.numberTypeNode)('u8'), arraySize);
            return (0, visitors_1.visit)(arrayType, self);
        },
        visitNumberType(numberType) {
            if (numberType.endian === 'le') {
                return {
                    type: numberType.format,
                    imports: new RustImportMap_1.RustImportMap(),
                    nestedStructs: [],
                };
            }
            // TODO: Add to the Rust validator.
            throw new Error('Number endianness not supported by Borsh');
        },
        visitPublicKeyType() {
            return {
                type: 'Pubkey',
                imports: new RustImportMap_1.RustImportMap().add('solana_program::pubkey::Pubkey'),
                nestedStructs: [],
            };
        },
        visitStringType() {
            if (!parentSize) {
                return {
                    type: `&str`,
                    imports: new RustImportMap_1.RustImportMap(),
                    nestedStructs: [],
                };
            }
            if (typeof parentSize === 'number') {
                return {
                    type: `[u8; ${parentSize}]`,
                    imports: new RustImportMap_1.RustImportMap(),
                    nestedStructs: [],
                };
            }
            if ((0, nodes_1.isNode)(parentSize, 'numberTypeNode') &&
                parentSize.endian === 'le') {
                switch (parentSize.format) {
                    case 'u32':
                        return {
                            type: 'String',
                            imports: new RustImportMap_1.RustImportMap(),
                            nestedStructs: [],
                        };
                    case 'u8':
                    case 'u16':
                    case 'u64': {
                        const prefix = parentSize.format.toUpperCase();
                        return {
                            type: `${prefix}PrefixString`,
                            imports: new RustImportMap_1.RustImportMap().add(`kaigan::types::${prefix}PrefixString`),
                            nestedStructs: [],
                        };
                    }
                    default:
                        throw new Error(`'String size not supported: ${parentSize.format}`);
                }
            }
            // TODO: Add to the Rust validator.
            throw new Error('String size not supported by Borsh');
        },
        visitFixedSizeType(fixedSizeType, { self }) {
            parentSize = fixedSizeType.size;
            const manifest = (0, visitors_1.visit)(fixedSizeType.type, self);
            parentSize = null;
            return manifest;
        },
        visitSizePrefixType(sizePrefixType, { self }) {
            parentSize = (0, nodes_1.resolveNestedTypeNode)(sizePrefixType.prefix);
            const manifest = (0, visitors_1.visit)(sizePrefixType.type, self);
            parentSize = null;
            return manifest;
        },
    }));
    return {
        ...visitor,
        setParentName: (name) => {
            parentName = name;
        },
        setNestedStruct: (value) => {
            nestedStruct = value;
        },
    };
}
exports.getTypeManifestVisitor = getTypeManifestVisitor;
function mergeManifests(manifests) {
    return {
        imports: new RustImportMap_1.RustImportMap().mergeWith(...manifests.map((td) => td.imports)),
        nestedStructs: manifests.flatMap((m) => m.nestedStructs),
    };
}
//# sourceMappingURL=getTypeManifestVisitor.js.map