"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.renderValueNodeVisitor = exports.renderValueNode = void 0;
const nodes_1 = require("../../nodes");
const shared_1 = require("../../shared");
const visitors_1 = require("../../visitors");
const RustImportMap_1 = require("./RustImportMap");
function renderValueNode(value, useStr = false) {
    return (0, visitors_1.visit)(value, renderValueNodeVisitor(useStr));
}
exports.renderValueNode = renderValueNode;
function renderValueNodeVisitor(useStr = false) {
    return {
        visitArrayValue(node) {
            const list = node.items.map((v) => (0, visitors_1.visit)(v, this));
            return {
                imports: new RustImportMap_1.RustImportMap().mergeWith(...list.map((c) => c.imports)),
                render: `[${list.map((c) => c.render).join(', ')}]`,
            };
        },
        visitBooleanValue(node) {
            return {
                imports: new RustImportMap_1.RustImportMap(),
                render: JSON.stringify(node.boolean),
            };
        },
        visitBytesValue(node) {
            const bytes = (0, nodes_1.getBytesFromBytesValueNode)(node);
            const numbers = Array.from(bytes).map(nodes_1.numberValueNode);
            return (0, visitors_1.visit)((0, nodes_1.arrayValueNode)(numbers), this);
        },
        visitConstantValue(node) {
            if ((0, nodes_1.isNode)(node.value, 'bytesValueNode')) {
                return (0, visitors_1.visit)(node.value, this);
            }
            if ((0, nodes_1.isNode)(node.type, 'stringTypeNode') &&
                (0, nodes_1.isNode)(node.value, 'stringValueNode')) {
                return (0, visitors_1.visit)((0, nodes_1.bytesValueNode)(node.type.encoding, node.value.string), this);
            }
            if ((0, nodes_1.isNode)(node.type, 'numberTypeNode') &&
                (0, nodes_1.isNode)(node.value, 'numberValueNode')) {
                const numberManifest = (0, visitors_1.visit)(node.value, this);
                const { format, endian } = node.type;
                const byteFunction = endian === 'le' ? 'to_le_bytes' : 'to_be_bytes';
                numberManifest.render = `${numberManifest.render}${format}.${byteFunction}()`;
                return numberManifest;
            }
            throw new Error('Unsupported constant value type.');
        },
        visitEnumValue(node) {
            var _a;
            const imports = new RustImportMap_1.RustImportMap();
            const enumName = (0, shared_1.pascalCase)(node.enum.name);
            const variantName = (0, shared_1.pascalCase)(node.variant);
            const importFrom = (_a = node.enum.importFrom) !== null && _a !== void 0 ? _a : 'generatedTypes';
            imports.add(`${importFrom}::${enumName}`);
            if (!node.value) {
                return { imports, render: `${enumName}::${variantName}` };
            }
            const enumValue = (0, visitors_1.visit)(node.value, this);
            const fields = enumValue.render;
            return {
                imports: imports.mergeWith(enumValue.imports),
                render: `${enumName}::${variantName} ${fields}`,
            };
        },
        visitMapValue(node) {
            const map = node.entries.map((entry) => (0, visitors_1.visit)(entry, this));
            const imports = new RustImportMap_1.RustImportMap().add('std::collection::HashMap');
            return {
                imports: imports.mergeWith(...map.map((c) => c.imports)),
                render: `HashMap::from([${map.map((c) => c.render).join(', ')}])`,
            };
        },
        visitMapEntryValue(node) {
            const mapKey = (0, visitors_1.visit)(node.key, this);
            const mapValue = (0, visitors_1.visit)(node.value, this);
            return {
                imports: mapKey.imports.mergeWith(mapValue.imports),
                render: `[${mapKey.render}, ${mapValue.render}]`,
            };
        },
        visitNoneValue() {
            return {
                imports: new RustImportMap_1.RustImportMap(),
                render: 'None',
            };
        },
        visitNumberValue(node) {
            return {
                imports: new RustImportMap_1.RustImportMap(),
                render: node.number.toString(),
            };
        },
        visitPublicKeyValue(node) {
            return {
                imports: new RustImportMap_1.RustImportMap().add('solana_program::pubkey'),
                render: `pubkey!("${node.publicKey}")`,
            };
        },
        visitSetValue(node) {
            const set = node.items.map((v) => (0, visitors_1.visit)(v, this));
            const imports = new RustImportMap_1.RustImportMap().add('std::collection::HashSet');
            return {
                imports: imports.mergeWith(...set.map((c) => c.imports)),
                render: `HashSet::from([${set.map((c) => c.render).join(', ')}])`,
            };
        },
        visitSomeValue(node) {
            const child = (0, visitors_1.visit)(node.value, this);
            return {
                ...child,
                render: `Some(${child.render})`,
            };
        },
        visitStringValue(node) {
            return {
                imports: new RustImportMap_1.RustImportMap(),
                render: useStr
                    ? `${JSON.stringify(node.string)}`
                    : `String::from(${JSON.stringify(node.string)})`,
            };
        },
        visitStructValue(node) {
            const struct = node.fields.map((field) => (0, visitors_1.visit)(field, this));
            return {
                imports: new RustImportMap_1.RustImportMap().mergeWith(...struct.map((c) => c.imports)),
                render: `{ ${struct.map((c) => c.render).join(', ')} }`,
            };
        },
        visitStructFieldValue(node) {
            const structValue = (0, visitors_1.visit)(node.value, this);
            return {
                imports: structValue.imports,
                render: `${node.name}: ${structValue.render}`,
            };
        },
        visitTupleValue(node) {
            const tuple = node.items.map((v) => (0, visitors_1.visit)(v, this));
            return {
                imports: new RustImportMap_1.RustImportMap().mergeWith(...tuple.map((c) => c.imports)),
                render: `(${tuple.map((c) => c.render).join(', ')})`,
            };
        },
    };
}
exports.renderValueNodeVisitor = renderValueNodeVisitor;
//# sourceMappingURL=renderValueNodeVisitor.js.map