"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getDefaultValidatorBagVisitor = void 0;
const nodes_1 = require("../nodes");
const shared_1 = require("../shared");
const extendVisitor_1 = require("./extendVisitor");
const getResolvedInstructionInputsVisitor_1 = require("./getResolvedInstructionInputsVisitor");
const mergeVisitor_1 = require("./mergeVisitor");
const recordLinkablesVisitor_1 = require("./recordLinkablesVisitor");
const recordNodeStackVisitor_1 = require("./recordNodeStackVisitor");
const visitor_1 = require("./visitor");
function getDefaultValidatorBagVisitor() {
    const linkables = new shared_1.LinkableDictionary();
    const stack = new shared_1.NodeStack();
    return (0, shared_1.pipe)((0, mergeVisitor_1.mergeVisitor)(() => new shared_1.ValidatorBag(), (_, bags) => new shared_1.ValidatorBag().mergeWith(bags)), (v) => (0, recordLinkablesVisitor_1.recordLinkablesVisitor)(v, linkables), (v) => (0, recordNodeStackVisitor_1.recordNodeStackVisitor)(v, stack), (v) => (0, extendVisitor_1.extendVisitor)(v, {
        visitProgram(node, { next }) {
            const bag = new shared_1.ValidatorBag();
            if (!node.name) {
                bag.error('Program has no name.', node, stack);
            }
            if (!node.publicKey) {
                bag.error('Program has no public key.', node, stack);
            }
            if (!node.version) {
                bag.warn('Program has no version.', node, stack);
            }
            if (!node.origin) {
                bag.info('Program has no origin.', node, stack);
            }
            return bag.mergeWith([next(node)]);
        },
        visitAccount(node, { next }) {
            const bag = new shared_1.ValidatorBag();
            if (!node.name) {
                bag.error('Account has no name.', node, stack);
            }
            return bag.mergeWith([next(node)]);
        },
        visitInstruction(node, { next }) {
            var _a;
            const bag = new shared_1.ValidatorBag();
            if (!node.name) {
                bag.error('Instruction has no name.', node, stack);
            }
            // Check for duplicate account names.
            const accountNameHistogram = new Map();
            node.accounts.forEach((account) => {
                var _a;
                if (!account.name) {
                    bag.error('Instruction account has no name.', node, stack);
                    return;
                }
                const count = ((_a = accountNameHistogram.get(account.name)) !== null && _a !== void 0 ? _a : 0) + 1;
                accountNameHistogram.set(account.name, count);
                // Only throw an error once per duplicated names.
                if (count === 2) {
                    bag.error(`Account name "${account.name}" is not unique in instruction "${node.name}".`, node, stack);
                }
            });
            // Check for cyclic dependencies in account defaults.
            const cyclicCheckVisitor = (0, getResolvedInstructionInputsVisitor_1.getResolvedInstructionInputsVisitor)();
            try {
                (0, visitor_1.visit)(node, cyclicCheckVisitor);
            }
            catch (error) {
                bag.error(error.message, node, stack);
            }
            // Check args.
            const names = [
                ...node.arguments.map(({ name }) => (0, shared_1.mainCase)(name)),
                ...((_a = node.extraArguments) !== null && _a !== void 0 ? _a : []).map(({ name }) => (0, shared_1.mainCase)(name)),
            ];
            const duplicates = names.filter((e, i, a) => a.indexOf(e) !== i);
            const uniqueDuplicates = [...new Set(duplicates)];
            const hasConflictingNames = uniqueDuplicates.length > 0;
            if (hasConflictingNames) {
                bag.error(`The names of the following instruction arguments are conflicting: ` +
                    `[${uniqueDuplicates.join(', ')}].`, node, stack);
            }
            // Check arg defaults.
            (0, nodes_1.getAllInstructionArguments)(node).forEach((argument) => {
                const { defaultValue } = argument;
                if ((0, nodes_1.isNode)(defaultValue, 'accountBumpValueNode')) {
                    const defaultAccount = node.accounts.find((account) => account.name === defaultValue.name);
                    if (defaultAccount && defaultAccount.isSigner !== false) {
                        bag.error(`Argument ${argument.name} cannot default to the bump attribute of ` +
                            `the [${defaultValue.name}] account as it may be a Signer.`, node, stack);
                    }
                }
            });
            return bag.mergeWith([next(node)]);
        },
        visitDefinedType(node, { next }) {
            const bag = new shared_1.ValidatorBag();
            if (!node.name) {
                bag.error('Defined type has no name.', node, stack);
            }
            return bag.mergeWith([next(node)]);
        },
        visitError(node, { next }) {
            const bag = new shared_1.ValidatorBag();
            if (!node.name) {
                bag.error('Error has no name.', node, stack);
            }
            if (typeof node.code !== 'number') {
                bag.error('Error has no code.', node, stack);
            }
            if (!node.message) {
                bag.warn('Error has no message.', node, stack);
            }
            return bag.mergeWith([next(node)]);
        },
        visitDefinedTypeLink(node, { next }) {
            const bag = new shared_1.ValidatorBag();
            if (!node.name) {
                bag.error('Pointing to a defined type with no name.', node, stack);
            }
            else if (!node.importFrom && !linkables.has(node)) {
                bag.error(`Pointing to a missing defined type named "${node.name}"`, node, stack);
            }
            return bag.mergeWith([next(node)]);
        },
        visitEnumType(node, { next }) {
            const bag = new shared_1.ValidatorBag();
            if (node.variants.length === 0) {
                bag.warn('Enum has no variants.', node, stack);
            }
            node.variants.forEach((variant) => {
                if (!variant.name) {
                    bag.error('Enum variant has no name.', node, stack);
                }
            });
            return bag.mergeWith([next(node)]);
        },
        visitEnumEmptyVariantType(node, { next }) {
            const bag = new shared_1.ValidatorBag();
            if (!node.name) {
                bag.error('Enum variant has no name.', node, stack);
            }
            return bag.mergeWith([next(node)]);
        },
        visitEnumStructVariantType(node, { next }) {
            const bag = new shared_1.ValidatorBag();
            if (!node.name) {
                bag.error('Enum variant has no name.', node, stack);
            }
            return bag.mergeWith([next(node)]);
        },
        visitEnumTupleVariantType(node, { next }) {
            const bag = new shared_1.ValidatorBag();
            if (!node.name) {
                bag.error('Enum variant has no name.', node, stack);
            }
            return bag.mergeWith([next(node)]);
        },
        visitStructType(node, { next }) {
            const bag = new shared_1.ValidatorBag();
            // Check for duplicate field names.
            const fieldNameHistogram = new Map();
            node.fields.forEach((field) => {
                var _a;
                if (!field.name)
                    return; // Handled by TypeStructField
                const count = ((_a = fieldNameHistogram.get(field.name)) !== null && _a !== void 0 ? _a : 0) + 1;
                fieldNameHistogram.set(field.name, count);
                // Only throw an error once per duplicated names.
                if (count === 2) {
                    bag.error(`Struct field name "${field.name}" is not unique.`, field, stack);
                }
            });
            return bag.mergeWith([next(node)]);
        },
        visitStructFieldType(node, { next }) {
            const bag = new shared_1.ValidatorBag();
            if (!node.name) {
                bag.error('Struct field has no name.', node, stack);
            }
            return bag.mergeWith([next(node)]);
        },
        visitTupleType(node, { next }) {
            const bag = new shared_1.ValidatorBag();
            if (node.items.length === 0) {
                bag.warn('Tuple has no items.', node, stack);
            }
            return bag.mergeWith([next(node)]);
        },
    }));
}
exports.getDefaultValidatorBagVisitor = getDefaultValidatorBagVisitor;
//# sourceMappingURL=getDefaultValidatorBagVisitor.js.map