"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getResolvedInstructionInputsVisitor = void 0;
const nodes_1 = require("../nodes");
const singleNodeVisitor_1 = require("./singleNodeVisitor");
function getResolvedInstructionInputsVisitor(options = {}) {
    var _a;
    const includeDataArgumentValueNodes = (_a = options.includeDataArgumentValueNodes) !== null && _a !== void 0 ? _a : false;
    let stack = [];
    let resolved = [];
    let visitedAccounts = new Map();
    let visitedArgs = new Map();
    function resolveInstructionInput(instruction, input) {
        // Ensure we don't visit the same input twice.
        if (((0, nodes_1.isNode)(input, 'instructionAccountNode') &&
            visitedAccounts.has(input.name)) ||
            ((0, nodes_1.isNode)(input, 'instructionArgumentNode') && visitedArgs.has(input.name))) {
            return;
        }
        // Ensure we don't have a circular dependency.
        const isCircular = stack.some(({ kind, name }) => kind === input.kind && name === input.name);
        if (isCircular) {
            const cycle = [...stack.map(({ name }) => name), input.name].join(' -> ');
            const error = `Circular dependency detected in the accounts and args of ` +
                `the "${instruction.name}" instruction. ` +
                `Got the following dependency cycle: ${cycle}.`;
            throw new Error(error);
        }
        // Resolve whilst keeping track of the stack.
        stack.push(input);
        const localResolved = input.kind === 'instructionAccountNode'
            ? resolveInstructionAccount(instruction, input)
            : resolveInstructionArgument(instruction, input);
        stack.pop();
        // Store the resolved input.
        resolved.push(localResolved);
        if (localResolved.kind === 'instructionAccountNode') {
            visitedAccounts.set(input.name, localResolved);
        }
        else {
            visitedArgs.set(input.name, localResolved);
        }
    }
    function resolveInstructionAccount(instruction, account) {
        var _a;
        // Find and visit dependencies first.
        const dependsOn = getInstructionDependencies(account);
        resolveInstructionDependencies(instruction, account, dependsOn);
        const localResolved = {
            ...account,
            isPda: (0, nodes_1.getAllInstructionArguments)(instruction).some((argument) => (0, nodes_1.isNode)(argument.defaultValue, 'accountBumpValueNode') &&
                argument.defaultValue.name === account.name),
            dependsOn,
            resolvedIsSigner: account.isSigner,
            resolvedIsOptional: account.isOptional,
        };
        switch ((_a = localResolved.defaultValue) === null || _a === void 0 ? void 0 : _a.kind) {
            case 'accountValueNode':
                const defaultAccount = visitedAccounts.get(localResolved.defaultValue.name);
                const resolvedIsPublicKey = account.isSigner === false && defaultAccount.isSigner === false;
                const resolvedIsSigner = account.isSigner === true && defaultAccount.isSigner === true;
                const resolvedIsOptionalSigner = !resolvedIsPublicKey && !resolvedIsSigner;
                localResolved.resolvedIsSigner = resolvedIsOptionalSigner
                    ? 'either'
                    : resolvedIsSigner;
                localResolved.resolvedIsOptional = defaultAccount.isOptional;
                break;
            case 'publicKeyValueNode':
            case 'programLinkNode':
            case 'programIdValueNode':
                localResolved.resolvedIsSigner =
                    account.isSigner === false ? false : 'either';
                localResolved.resolvedIsOptional = false;
                break;
            case 'pdaValueNode':
                localResolved.resolvedIsSigner =
                    account.isSigner === false ? false : 'either';
                localResolved.resolvedIsOptional = false;
                const { seeds } = localResolved.defaultValue;
                seeds.forEach((seed) => {
                    if (!(0, nodes_1.isNode)(seed.value, 'accountValueNode'))
                        return;
                    const dependency = visitedAccounts.get(seed.value.name);
                    if (dependency.resolvedIsOptional) {
                        const error = `Cannot use optional account "${seed.value.name}" as the "${seed.name}" PDA seed ` +
                            `for the "${account.name}" account of the "${instruction.name}" instruction.`;
                        throw new Error(error);
                    }
                });
                break;
            case 'identityValueNode':
            case 'payerValueNode':
            case 'resolverValueNode':
                localResolved.resolvedIsOptional = false;
                break;
            default:
                break;
        }
        return localResolved;
    }
    function resolveInstructionArgument(instruction, argument) {
        // Find and visit dependencies first.
        const dependsOn = getInstructionDependencies(argument);
        resolveInstructionDependencies(instruction, argument, dependsOn);
        return { ...argument, dependsOn };
    }
    function resolveInstructionDependencies(instruction, parent, dependencies) {
        dependencies.forEach((dependency) => {
            let input = null;
            if ((0, nodes_1.isNode)(dependency, 'accountValueNode')) {
                const dependencyAccount = instruction.accounts.find((a) => a.name === dependency.name);
                if (!dependencyAccount) {
                    const error = `Account "${dependency.name}" is not a valid dependency of ${parent.kind} ` +
                        `"${parent.name}" in the "${instruction.name}" instruction.`;
                    throw new Error(error);
                }
                input = { ...dependencyAccount };
            }
            else if ((0, nodes_1.isNode)(dependency, 'argumentValueNode')) {
                const dependencyArgument = (0, nodes_1.getAllInstructionArguments)(instruction).find((a) => a.name === dependency.name);
                if (!dependencyArgument) {
                    const error = `Argument "${dependency.name}" is not a valid dependency of ${parent.kind} ` +
                        `"${parent.name}" in the "${instruction.name}" instruction.`;
                    throw new Error(error);
                }
                input = { ...dependencyArgument };
            }
            if (input) {
                resolveInstructionInput(instruction, input);
            }
        });
    }
    function getInstructionDependencies(input) {
        var _a;
        if (!input.defaultValue)
            return [];
        const getNestedDependencies = (defaultValue) => {
            if (!defaultValue)
                return [];
            return getInstructionDependencies({ ...input, defaultValue });
        };
        if ((0, nodes_1.isNode)(input.defaultValue, ['accountValueNode', 'accountBumpValueNode'])) {
            return [(0, nodes_1.accountValueNode)(input.defaultValue.name)];
        }
        if ((0, nodes_1.isNode)(input.defaultValue, 'pdaValueNode')) {
            const dependencies = new Map();
            input.defaultValue.seeds.forEach((seed) => {
                if ((0, nodes_1.isNode)(seed.value, ['accountValueNode', 'argumentValueNode'])) {
                    dependencies.set(seed.value.name, { ...seed.value });
                }
            });
            return [...dependencies.values()];
        }
        if ((0, nodes_1.isNode)(input.defaultValue, 'resolverValueNode')) {
            return (_a = input.defaultValue.dependsOn) !== null && _a !== void 0 ? _a : [];
        }
        if ((0, nodes_1.isNode)(input.defaultValue, 'conditionalValueNode')) {
            return [
                ...getNestedDependencies(input.defaultValue.condition),
                ...getNestedDependencies(input.defaultValue.ifTrue),
                ...getNestedDependencies(input.defaultValue.ifFalse),
            ];
        }
        return [];
    }
    return (0, singleNodeVisitor_1.singleNodeVisitor)('instructionNode', (node) => {
        var _a;
        // Ensure we always start with a clean slate.
        stack = [];
        resolved = [];
        visitedAccounts = new Map();
        visitedArgs = new Map();
        const inputs = [
            ...node.accounts,
            ...node.arguments.filter((a) => {
                if (includeDataArgumentValueNodes)
                    return a.defaultValue;
                return a.defaultValue && !(0, nodes_1.isNode)(a.defaultValue, nodes_1.VALUE_NODES);
            }),
            ...((_a = node.extraArguments) !== null && _a !== void 0 ? _a : []).filter((a) => a.defaultValue),
        ];
        // Visit all instruction accounts.
        inputs.forEach((input) => {
            resolveInstructionInput(node, input);
        });
        return resolved;
    });
}
exports.getResolvedInstructionInputsVisitor = getResolvedInstructionInputsVisitor;
//# sourceMappingURL=getResolvedInstructionInputsVisitor.js.map